@@ -35,15 +35,19 @@ Library for parsing package.xml and providing an object
 representation.
 """
 
+from __future__ import print_function
+
 import os
 import re
 import sys
 import xml.dom.minidom as dom
 
+PACKAGE_MANIFEST_FILENAME = 'package.xml'
+
 
 class Package(object):
     """
-    Object representation of a ``package.xml`` file
+    Object representation of a package manifest file
     """
     __slots__ = [
         'package_format',
@@ -105,44 +109,53 @@ class Package(object):
         :param package: Package to check
         :raises InvalidPackage: in case validation fails
         """
+        errors = []
         if self.package_format:
             if not re.match('^[1-9][0-9]*$', str(self.package_format)):
-                raise InvalidPackage('The "format" attribute of the package must contain a positive integer if present')
+                errors.append('The "format" attribute of the package must contain a positive integer if present')
 
         if not self.name:
-            raise InvalidPackage('Package name must not be empty')
+            errors.append('Package name must not be empty')
         # accepting upper case letters and hyphens only for backward compatibility
         if not re.match('^[a-zA-Z0-9][a-zA-Z0-9_-]*$', self.name):
-            raise InvalidPackage('Package name "%s" does not follow naming conventions' % self.name)
-        if not re.match('^[a-z][a-z0-9_]*$', self.name):
-            sys.stderr.write('WARNING: Package name "%s" does not follow the naming conventions. It should start with a lower case letter and only contain lower case letters, digits and underscores.\n' % self.name)
+            errors.append('Package name "%s" does not follow naming conventions' % self.name)
+        elif not re.match('^[a-z][a-z0-9_]*$', self.name):
+            print('WARNING: Package name "%s" does not follow the naming conventions. It should start with a lower case letter and only contain lower case letters, digits and underscores.' % self.name, file=sys.stderr)
 
         if not self.version:
-            raise InvalidPackage('Package version must not be empty')
-        if not re.match('^[0-9]+\.[0-9_]+\.[0-9_]+$', self.version):
-            raise InvalidPackage('Package version "%s" does not follow version conventions' % self.version)
+            errors.append('Package version must not be empty')
+        elif not re.match('^[0-9]+\.[0-9_]+\.[0-9_]+$', self.version):
+            errors.append('Package version "%s" does not follow version conventions' % self.version)
 
         if not self.description:
-            raise InvalidPackage('Package description must not be empty')
+            errors.append('Package description must not be empty')
 
         if not self.maintainers:
-            raise InvalidPackage('Package must declare at least one maintainer')
+            errors.append('Package must declare at least one maintainer')
         for maintainer in self.maintainers:
-            maintainer.validate()
+            try:
+                maintainer.validate()
+            except InvalidPackage as e:
+                errors.append(str(e))
             if not maintainer.email:
-                raise InvalidPackage('Maintainers must have an email address')
+                errors.append('Maintainers must have an email address')
 
         if not self.licenses:
-            raise InvalidPackage('The manifest must contain at least one "license" tag')
+            errors.append('The package node must contain at least one "license" tag')
 
         if self.authors is not None:
             for author in self.authors:
-                author.validate()
+                try:
+                    author.validate()
+                except InvalidPackage as e:
+                    errors.append(str(e))
 
         for dep_type, depends in {'build': self.build_depends, 'buildtool': self.buildtool_depends, 'run': self.run_depends, 'test': self.test_depends}.items():
             for depend in depends:
                 if depend.name == self.name:
-                    raise InvalidPackage('The manifest must not "%s_depend" on a package with the same name as this package' % dep_type)
+                    errors.append('The package must not "%s_depend" on a package with the same name as this package' % dep_type)
+        if errors:
+            raise InvalidPackage('\n'.join(errors))
 
 
 class Dependency(object):
@@ -170,6 +183,16 @@ class Export(object):
         self.attributes = {}
         self.content = content
 
+    def __str__(self):
+        txt = '<%s' % self.tagname
+        for key in sorted(self.attributes.keys()):
+            txt += ' %s="%s"' % (key, self.attributes[key])
+        if self.content:
+            txt += '>%s</%s>' % (self.content, self.tagname)
+        else:
+            txt += '/>'
+        return txt
+
 
 class Person(object):
     __slots__ = ['name', 'email']
@@ -275,11 +298,11 @@ def parse_package(path):
     if os.path.isfile(path):
         filename = path
     elif os.path.isdir(path):
-        filename = os.path.join(path, 'package.xml')
+        filename = os.path.join(path, PACKAGE_MANIFEST_FILENAME)
         if not os.path.isfile(filename):
-            raise IOError('Directory "%s" does not contain a "package.xml"' % (path))
+            raise IOError('Directory "%s" does not contain a "%s"' % (path, PACKAGE_MANIFEST_FILENAME))
     else:
-        raise IOError('Path "%s" is neither a directory containing a "package.xml" file nor a file' % (path))
+        raise IOError('Path "%s" is neither a directory containing a "%s" file nor a file' % (path, PACKAGE_MANIFEST_FILENAME))
 
     with open(filename, 'r') as f:
         try:
@@ -373,7 +396,48 @@ def parse_package_string(data, filename=None):
                 export.attributes[str(key)] = str(value)
             exports.append(export)
         pkg.exports = exports
+
+    errors = []
+    # verify that no unsupported tags and attributes are present
+    unknown_root_attributes = [attr for attr in root.attributes.keys() if str(attr) != 'format']
+    if unknown_root_attributes:
+        errors.append('The "package" tag must not have the following attributes: %s' % ', '.join(unknown_root_attributes))
+    depend_attributes = ['version_lt', 'version_lte', 'version_eq', 'version_gte', 'version_gt']
+    known = {
+        'name': [],
+        'version': ['abi'],
+        'description': [],
+        'maintainer': ['email'],
+        'license': [],
+        'url': ['type'],
+        'author': ['email'],
+        'build_depend': depend_attributes,
+        'buildtool_depend': depend_attributes,
+        'run_depend': depend_attributes,
+        'test_depend': depend_attributes,
+        'conflict_depend': depend_attributes,
+        'replace_depend': depend_attributes,
+        'export': [],
+    }
+    nodes = [n for n in root.childNodes if n.nodeType == n.ELEMENT_NODE]
+    unknown_tags = set([n.tagName for n in nodes if n.tagName not in known.keys()])
+    if unknown_tags:
+        errors.append('The manifest must not contain the following tags: %s' % ', '.join(unknown_tags))
+    for node in [n for n in nodes if n.tagName in known.keys()]:
+        unknown_attrs = [str(attr) for attr in node.attributes.keys() if str(attr) not in known[node.tagName]]
+        if unknown_attrs:
+            errors.append('The "%s" tag must not have the following attributes: %s' % (node.tagName, ', '.join(unknown_attrs)))
+        if node.tagName not in ['description', 'export']:
+            subnodes = [n for n in node.childNodes if n.nodeType == n.ELEMENT_NODE]
+            if subnodes:
+                errors.append('The "%s" tag must not contain the following children: %s' % (node.tagName, ', '.join([n.tagName for n in subnodes])))
+    if errors:
+        # for now only output a warning instead of raising an exception
+        #raise InvalidPackage('\n'.join(errors))
+        print('WARNING:%s' % ''.join(['\n- %s' % e for e in errors]), file=sys.stderr)
+
     pkg.validate()
+
     return pkg
 
 
